# -*- coding: utf-8 -*-
"""
@author: Md Tauhidul Islam, Postdoc, Xing Lab,
Department of Radiation Oncology, Stanford University
 
"""

import numpy as np
from scipy.spatial.distance import cdist  
from scipy.sparse import issparse  
from sklearn.cluster import kmeans_plusplus


class groupCentroidOPT:
    """ gc = groupCentroidOPT( X, k= or centres=, ... ) applies the first
    optimization to group the data and compute a number of centroids
        in: either initial centres= for minimizeCost()
            or k= [nsample=] for runCostMinimization
        out: gc.centres, gc.Xtocentre, gc.distances
        iterator:
            for jcentre, J in gc
                clustercentre = centres[jcentre]
                J indexes e.g. X[J], classes[J]
    """
    def __init__( self, X, k=0, centres=None, nsample=0, **kwargs ):
        self.X = X
        if centres is None:
            self.centres, self.Xtocentre, self.distances = runCostMinimization(
                X, k=k, nsample=nsample, **kwargs )
        else:
            self.centres, self.Xtocentre, self.distances = minimizeCost(
                X, centres, **kwargs )

    def __iter__(self):
        for jc in range(len(self.centres)):
            yield jc, (self.Xtocentre == jc)

def minimizeCost( X, centres, delta=.001, maxiter=10, metric="euclidean", p=2, verbose=1 ):
    """ centres, Xtocentre, distances = minimizeCost( X, initial centres ... )
    in:
        X N x dim  may be sparse
        centres k x dim: initial centres, e.g. random.sample( X, k )
        delta: relative error, iterate until the average distance to centres
            is within delta of the previous average distance
        maxiter
        metric: any of the 20-odd in scipy.spatial.distance
            "chebyshev" = max, "cityblock" = L1, "minkowski" with p=
            or a function( Xvec, centrevec ), e.g. Lqmetric below
        p: for minkowski metric -- local mod cdist for 0 < p < 1 too
        verbose: 0 silent, 2 prints running distances
    out:
        centres, k x dim
        Xtocentre: each X -> its nearest centre, ints N -> k
        distances, N
    """
    if not issparse(X):
        X = np.asanyarray(X)  # ?
    centres = centres.todense() if issparse(centres) \
        else centres.copy()
    N, dim = X.shape
    k, cdim = centres.shape
    if dim != cdim:
        raise ValueError( "X %s and centres %s must have the same number of columns" % (
            X.shape, centres.shape ))

    allx = np.arange(N)
    prevdist = 0
    for jiter in range( 1, maxiter+1 ):
        D = cdist_sparse( X, centres, metric=metric)  # |X| x |centres|
        xtoc = D.argmin(axis=1)  # X -> nearest centre
        distances = D[allx,xtoc]
        avdist = distances.mean()  

        if (1 - delta) * prevdist <= avdist <= prevdist \
        or jiter == maxiter:
            break
        prevdist = avdist
        for jc in range(k):  # (1 pass in C)
            c = np.where( xtoc == jc )[0]
            if len(c) > 0:
                centres[jc] = X[c].mean( axis=0 )

    if verbose >= 2:
        r50 = np.zeros(k)
        r90 = np.zeros(k)
        for j in range(k):
            dist = distances[ xtoc == j ]
            if len(dist) > 0:
                r50[j], r90[j] = np.percentile( dist, (50, 90) )

    return centres, xtoc, distances

def runCostMinimization( X, k, nsample=0, **kwargs ):
    """ Initialization of the centroids using kmeans++ and then applying 
    group centroid optimization starting from those centroids
    """
    samplecentres, indices = kmeans_plusplus(X, n_clusters=k,random_state=0)    
    return minimizeCost( X, samplecentres, **kwargs )

        
def cdist_sparse( X, Y, metric ):
    """ -> |X| x |Y| cdist array, any cdist metric
        X or Y may be sparse -- best csr
    """
    sxy = 2*issparse(X) + issparse(Y)
    if sxy == 0:
        return cdist( X, Y, metric )
    d = np.empty( (X.shape[0], Y.shape[0]), np.float64 )
    if sxy == 2:
        for j, x in enumerate(X):
            d[j] = cdist( x.todense(), Y, metric ) [0]
    elif sxy == 1:
        for k, y in enumerate(Y):
            d[:,k] = cdist( X, y.todense(), metric ) [0]
    else:
        for j, x in enumerate(X):
            for k, y in enumerate(Y):
                d[j,k] = cdist( x.todense(), y.todense(), metric ) [0]
    return d

def nearestcentres( X, centres, metric="euclidean", p=2 ):
    """ each X -> nearest centre, any metric
            euclidean2 (~ withinss) is more sensitive to outliers,
            cityblock (manhattan, L1) less sensitive
    """
    D = cdist( X, centres, metric=metric, p=p )  # |X| x |centres|
    return D.argmin(axis=1)

def Lqmetric( x, y=None, q=.5 ):
    return (np.abs(x - y) ** q) .mean() if y is not None \
        else (np.abs(x) ** q) .mean()



from keras.models import Sequential
from keras.layers import Conv2D, MaxPooling2D, Flatten, Dense
from lime import lime_image
import numpy as np
from tensorflow.keras.utils import to_categorical

def compute_genoSig(imageset,label, sig_class, epochs=100):
    # imageset contains the genomaps in shape (cellNum, rowNum, colNum, 1)
    # label: numpy array of shape (cellNum, 1) containing the cell labels
    # sig_class: numpy array containing the cell classes for which gene signatures will be computed

    n_clusters = len(np.unique(label))
    # Create the deep learning model
    model = Sequential()
    model.add(Conv2D(32, kernel_size=(3, 3), activation='relu', input_shape=(imageset.shape[1], imageset.shape[2], imageset.shape[3])))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(64, kernel_size=(3, 3), activation='relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Flatten())
    model.add(Dense(128, activation='relu'))
    model.add(Dense(n_clusters + 1, activation='softmax'))

    # Compile and train the model
    model.compile(loss='sparse_categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
    model.fit(imageset, label, batch_size=32, epochs=epochs)

    label = to_categorical(label)
    # Use LIME for interpretability
    explainer = lime_image.LimeImageExplainer(verbose = False)
    imgs=[]
    lbls=[]
    #sig_classARR = np.asarray([np.argmax(row) for row in sig_class])
    sig_classARR=sig_class
    labelARR = np.asarray([np.argmax(row) for row in label])
    for i in range(sig_classARR.shape[0]):  # Assuming you have 10 classes
        # Get images of class i
        #idx = [label == sig_class[i]]
        matching_indices_full = np.where(np.isin(labelARR, sig_classARR[i]))[0]
        img=imageset[matching_indices_full]
        imgs.append(img)
        lbl=label[matching_indices_full]
        lbls.append(lbl)

    # Convert the list to a numpy array

    imgs = np.vstack(imgs)
    lbls = np.vstack(lbls)

    catLime_images = []
    # Select a random image from the test set
    for i in range(imgs.shape[0]):

        image = imgs[i]
        ground_truth = np.argmax(lbls[i])
        # Generate an explanation using LIME
        explanation = explainer.explain_instance(image, model.predict, top_labels=n_clusters, hide_color=0, num_samples=100)
        
        # Get the explanation for the ground truth class
        lime_img, mask = explanation.get_image_and_mask(ground_truth, positive_only=False, num_features=imgs.shape[1]*imgs.shape[2], hide_rest=False)
        gray_image = rgb2gray(lime_img)
        catLime_images.append(gray_image)
    catLime_images = np.array(catLime_images)

    lblsARR = np.asarray([np.argmax(row) for row in lbls])
    mean_images = []
    for i in range(sig_classARR.shape[0]):
        matching_indices_full = np.where(np.isin(lblsARR, sig_classARR[i]))[0]
        class_i_images = catLime_images[matching_indices_full]
        mean_image = np.mean(class_i_images, axis=0)
        mean_images.append(mean_image)

    return mean_images


def rgb2gray(image):
    return np.dot(image[...,:3], [0.2989, 0.5870, 0.1140])

import numpy as np
import pandas as pd

def arrays_to_dataframe(arrays, strings):
    # Check if the number of arrays is even
    if len(arrays) % 2 != 0:
        raise ValueError("The number of arrays should be even.")

    # Check if the number of strings is half the number of arrays
    if len(strings) != len(arrays) // 2:
        raise ValueError("The number of strings should be half the number of arrays.")

    # Initialize an empty DataFrame
    df = pd.DataFrame()

    # For each pair of arrays
    for i in range(0, len(arrays), 2):
        # Get the corresponding string
        string = np.squeeze(strings[i // 2])

        # Add the two arrays to the DataFrame as columns with the same string
        df[str(string) + '_geneName'] = arrays[i]
        df[str(string) + '_importance'] = arrays[i + 1]

    return df

import pandas as pd
from sklearn.preprocessing import LabelEncoder

def genoSig(genoMaps,T,label,userPD,gene_names, epochs=100):
    genoMaps_3d = np.repeat(genoMaps, 3, axis=-1)

    # first, convert the strings to integer labels
    label_encoder = LabelEncoder()
    label_encoded = label_encoder.fit_transform(label)

    lc = np.array([])
    for i in range(0, userPD.shape[0]):
        first_occurrence = (label == userPD[i]).idxmax()
        lc = np.append(lc, label_encoded[first_occurrence[0]])

    lc = np.array(lc)

    n_clusters = len(np.unique(label))
    y_train = to_categorical(label_encoded)
    # meanI=compute_genoSig(X_train,y_train, [y_train[0],y_train[1]])
    meanI = compute_genoSig(genoMaps_3d, label_encoded, lc, epochs=epochs)



    result = pd.DataFrame()

    for ii in range(0, len(meanI)):
        meanIX = meanI[ii]
        Ivec = np.reshape(meanIX, (1, meanIX.shape[0] * meanIX.shape[1]), order='F').copy()
        gene_namesN = np.array(gene_names)
        row_indices = np.argmax(T, axis=0)
        gene_namesT = gene_namesN[row_indices]
        Ivec = np.squeeze(Ivec)
        indices = np.argsort(Ivec)[::-1]
        Ivec_sortedX = Ivec[indices]
        gene_namesTX = gene_namesT[indices]
        labelD = np.array(label)
        strings = [userPD[ii]]
        arrays = np.vstack((gene_namesTX, Ivec_sortedX))
        df = arrays_to_dataframe(arrays, strings)
        result = horizontal_concat = pd.concat([result, df], axis=1)

    return result



"""
Created on Sun Jul 16 21:27:17 2023
@author: Md Tauhidul Islam, Research Scientist, Dept. of radiation Oncology, Stanford University
"""

import phate
import numpy as np
from genomap.utils.class_discriminative_opt import ClassDiscriminative_OPT
import scipy
from genomap.utils.gTraj_utils import compute_cluster_distances, nearest_divisible_by_four
from tensorflow.keras.optimizers import Adam
from genomap.utils.ConvIDEC import ConvIDEC
from sklearn.feature_selection import VarianceThreshold
from genomap.genomap import construct_genomap
from genomap.utils.util_Sig import select_n_features

def apply_genoTraj(data, y_pred):
# data: input data
# y_pred: predicted pseudo labels
    Kdis=compute_cluster_distances(data,y_pred);

# Kdis = normalize(Kdis, axis=1, norm='l1')
# Kdis=1-Kdis
    n_clusters=len(np.unique(y_pred))
    num_comp=n_clusters-1
    # Second optimization (class-discriminative optimization)
    clf = ClassDiscriminative_OPT(n_components=num_comp)
    clf.fit(Kdis, y_pred)
    ccifOut = clf.transform(Kdis)
    ccifOutzc = scipy.stats.zscore(ccifOut, axis=0, ddof=1) # Z-score of the CCIF
    phate_op = phate.PHATE()
    data_phate = phate_op.fit_transform(ccifOutzc)
    return data_phate

def genoTraj(data,n_clusters = 33, colNum=32,rowNum=32,batch_size=64,verbose=1,
                    pretrain_epochs=100,maxiter=300):
# rowNum and colNum are the row and column numbers of constructed genomaps
# n_clusters: number of  pseudo-classes in the data, should be set as 17, 33 or 65
# batch_size: number of samples in each mini batch while training the deep neural network
# verbose: whether training progress will be shown or not
# pretrain_epochs: number of epoch for pre-training the CNN
# maxiter: number of epoch of fine-tuning training                    
# Construction of genomap    
    colNum=nearest_divisible_by_four(colNum)
    rowNum=nearest_divisible_by_four(rowNum)
    nump=rowNum*colNum 
    if nump<data.shape[1]:
        data,index=select_n_features(data,nump)
    genoMaps=construct_genomap(data,rowNum,colNum,epsilon=0.0,num_iter=200)

# Deep learning-based trajectory mapping
    optimizer = Adam()    
    model = ConvIDEC(input_shape=genoMaps.shape[1:], filters=[32, 64, 128, 32], n_clusters=n_clusters)
    model.compile(optimizer=optimizer, loss=['kld', 'mse'], loss_weights=[0.1, 1.0])
    pretrain_optimizer ='adam'
    update_interval=50
    model.pretrain(genoMaps, y=None, optimizer=pretrain_optimizer, epochs=pretrain_epochs, batch_size=batch_size,
                        verbose=verbose)

    y_pred = model.fit(genoMaps, y=None, maxiter=maxiter, batch_size=batch_size, update_interval=update_interval,
                       )
    y_pred = model.predict_labels(genoMaps) 

    dataX = scipy.stats.zscore(data, axis=0, ddof=1)
    outGenoTraj=apply_genoTraj(dataX,y_pred)
    return outGenoTraj 




"""
Created on Sun Jul 16 21:27:17 2023
@author: Md Tauhidul Islam, Research Scientist, Dept. of radiation Oncology, Stanford University
"""

import os

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
import torch.nn.functional as F


torch.manual_seed(0)
torch.use_deterministic_algorithms(True)
def findConv2dOutShape(hin,win,conv,pool=2):
    # get conv arguments
    kernel_size=conv.kernel_size
    stride=conv.stride
    padding=conv.padding
    dilation=conv.dilation

    hout=np.floor((hin+2*padding[0]-dilation[0]*(kernel_size[0]-1)-1)/stride[0]+1)
    wout=np.floor((win+2*padding[1]-dilation[1]*(kernel_size[1]-1)-1)/stride[1]+1)

    if pool:
        hout/=pool
        wout/=pool
    return int(hout),int(wout)


class genoNet(nn.Module):
    # Define the convolutional neural network architecture
    def __init__(self, input_shape, class_num):
        super(genoNet, self).__init__()
        
        
        input_dim=input_shape[2]
        Cin, Hin, Win = 1, input_dim, input_dim
        init_f = 8
        num_fc1 = 100
        
        self.conv1 = nn.Conv2d(Cin, init_f, kernel_size=3, padding=1)
        h,w=findConv2dOutShape(Hin,Win,self.conv1,pool=0)
        
        self.num_flatten=h*w*init_f
        self.fc1 = nn.Linear(self.num_flatten, num_fc1)
        self.fc2 = nn.Linear(num_fc1, class_num)
        self.dropout = nn.Dropout(0.25)
        
        
        self.softmax = nn.Softmax(dim=-1)
        
 
    def forward(self, x):
        x = F.relu(self.conv1(x)); 
        x = x.contiguous().view(-1, self.num_flatten)
        x = F.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.fc2(x)
        
        return F.log_softmax(x, dim=1)
    
    def forwardX(self, x):
        x = F.relu(self.conv1(x)); 
        x = x.contiguous().view(-1, self.num_flatten)
        x = F.relu(self.fc1(x))

        return x
    
def get_device():
    return _get_device()

def _get_device():
    # help function to detect whether the computer has a GPU
    device = 'cpu'
    return device

def fit(model, dataloader, epoch, criterion, optimizer, device, trainset):
    # fit one epoch
    #
    # model: the genoNet model
    # epoch: the current epoch number
    # criterion: the loss function 
    # optimizer: the optimizer (by default Adam)
    # device: the device that the model is running on
    # trainset: the training dataset
    
    model.train()
    running_loss = 0.0
    counter = 0
    
    for i, data in tqdm(enumerate(dataloader),
                        total=int(len(trainset)/dataloader.batch_size), disable=True):
        counter += 1
        img = data[0]
        img = img.to(device).float()
        
        label = data[1]
        label = label.to(device).long()
        optimizer.zero_grad()
        outputs = model(img)

        loss = criterion(outputs, label)
        
        loss.backward()
        optimizer.step()
        running_loss += loss.item()
    
    epoch_loss = running_loss / counter
    return epoch_loss

class geneDataset(Dataset):
    # Dataset for gene expression data
    # 
    # Attribute:
    #  data: rows denote cells and columns denote the genes. 
    def __init__(self,data,label):
        self.data = data
        self.label = label
        
    def __getitem__(self, index):
        return self.data[index,:,:,:], self.label[index]
    
    def __len__(self):
        return len(self.data)

def predict(model, dataloader, device):
    # predict the transformed data to calculate ideal distribution with trained model
    prediction_list = []
    model.eval()
    running_loss = 0.0
    counter = 0
    with torch.no_grad():
        for i, data in enumerate(dataloader):
            img = data[0]
            img = img.to(device).float()

            pred = model(img)
            prediction_list.append(pred.cpu().numpy())
    return np.concatenate(prediction_list)
    
def traingenoNet(data, labels, maxEPOCH, batchSize=16, verbose=False):
    # train the Sparse Autoencoder model
    #
    # data: training data, rows denote cells and columns denote the genes.
    # reduced_Dim: neurons in the hidden layer
    # maxEPOCH: max training EPOCH number
    # batchSize: batchSize for training
    # l2: the weight for l2 regularizer of all parameters
    # sparse_rho: the desired value for KL Divergence
    # sparse: the weight for sparse regularizer (KL Divergence loss)
    device = torch.device('cpu')
    
    trainset = geneDataset(data, labels)
    trainloader = DataLoader(trainset,
                             batch_size=batchSize,
                             shuffle=True)
    
    input_dim = data.shape[1:] # 33 x 33 x 1
    class_num, class_weight = np.unique(labels, return_counts=True)
    class_num = len(class_num)
    class_weight = torch.Tensor(np.sum(class_weight) / class_weight).to(device)
    model = genoNet(input_dim, class_num).to(device)
    
    criterion = nn.CrossEntropyLoss(weight=class_weight)
    optimizer = optim.AdamW(model.parameters(), lr=1e-3, weight_decay=1e-5)
    train_loss = []
    for epoch in range(maxEPOCH):
        train_epoch_loss = fit(model, trainloader, epoch,
                               criterion, optimizer,device, trainset)
        if verbose:
            if epoch % 10 == 0:
                print(str(epoch) + ': ' + str(train_epoch_loss))
        train_loss.append(train_epoch_loss)
    
    
    return model

def load_genoNet(input_shape, class_num, path):
    device = torch.device('cpu')
    model = genoNet(input_shape, class_num)
    model.load_state_dict(torch.load(path, map_location=device))
    model.eval()
    return model

def rescale(B):
    # Rescale a vector from 0 to 1
    Bc=(B-np.min(B))/(np.max(B)-np.min(B))
    return Bc




"""
Created on Sun Jul 16 21:27:17 2023
@author: Md Tauhidul Islam, Research Scientist, Dept. of radiation Oncology, Stanford University
"""

import numpy as np
import os
import torch
from torch.utils.data import DataLoader, Dataset
import scipy

from genomap.genomap import construct_genomap
from genomap.utils.util_Sig import select_n_features
import genomap.genoNet as gNet

def genoClassification(training_data, training_labels, test_data, rowNum=32, colNum=32, epoch=100):

    # training_data: numpy array in cellxgene shape
    # training_labels: numpy array 
    # test_data: numpy array in cellxgene shape
    # rowNum: row number of genomap
    # colNum: column number of genomap
    # epoch: epoch number for training
    
    data = np.concatenate((training_data, test_data), axis=0)
    # Construction of genomaps
    nump=rowNum*colNum 
    if nump<data.shape[1]:
        data,index=select_n_features(data,nump)
    
    dataNorm=scipy.stats.zscore(data,axis=0,ddof=1)
    genoMaps=construct_genomap(dataNorm,rowNum,colNum,epsilon=0.0,num_iter=200)    

    # Split the data for training and testing
    dataMat_CNNtrain = genoMaps[:training_labels.shape[0]] 
    dataMat_CNNtest = genoMaps[training_labels.shape[0]:]

    groundTruthTrain = training_labels
    classNum = len(np.unique(groundTruthTrain))

    # Preparation of training and testing data for PyTorch computation
    XTrain = dataMat_CNNtrain.transpose([0,3,1,2])
    XTest = dataMat_CNNtest.transpose([0,3,1,2])
    yTrain = groundTruthTrain

    # Train the network in PyTorch framework
    miniBatchSize = 128
    net = gNet.traingenoNet(XTrain, yTrain, maxEPOCH=epoch, batchSize=miniBatchSize, verbose=True)

    # # Process the test data
    yTest = np.random.rand(dataMat_CNNtest.shape[0])
    testset = gNet.geneDataset(XTest, yTest)
    testloader = gNet.DataLoader(testset, batch_size=miniBatchSize, shuffle=False)
    device = gNet.get_device()

    # Perform data classification
    prob_test = gNet.predict(net, testloader, device)
    pred_test = np.argmax(prob_test, axis=-1)
    
    return pred_test
    
    
    
    



import random
import numpy as np

def select_random_values(start, end, perc):
    total_values = end - start + 1
    num_values_to_select = int(total_values * perc)
    
    all_values = list(range(start, end + 1))
    selected_values = random.sample(all_values, num_values_to_select)
    remaining_values = list(set(all_values) - set(selected_values))
    
    return np.array(selected_values), np.array(remaining_values) 



# -*- coding: utf-8 -*-
"""
Created on Sun Jul 16 18:25:32 2023

@author: Windows
"""

import numpy as np
from genomap.genomapT import construct_genomap_returnT
# suppose you have a feature matrix X
# X = np.array(...)

def select_n_features(X,n):
# calculate the variance of each feature
    variances = np.var(X, axis=0)

# get the indices of the features sorted by variance (in descending order)
    indices = np.argsort(variances)[::-1]

# select the indices of the top 10 most variable features
    top_n_indices = indices[:n]

# select the top 10 most variable features
    X_top_n = X[:, top_n_indices]
    return X_top_n,top_n_indices


def createGenomap_for_sig(data,gene_names,rowNum=32,colNum=32):

    """
    Returns the constructed genomaps


    Parameters
    ----------
    data : ndarray, shape (cellNum, geneNum)
         gene expression data in cell X gene format. Each row corresponds
         to one cell, whereas each column represents one gene
    gene_names: numpy array, shape (1, geneNum)
        name of the genes corresponding to the columns
    rowNum : int, 
         number of rows in a genomap
    colNum : int,
         number of columns in a genomap

    Returns
    -------
    genomaps : ndarray, shape (rowNum, colNum, zAxisDimension, cell number)
           genomaps are formed for each cell. zAxisDimension is more than
           1 when 3D genomaps are created. 
    T : transfer function of the genomap construction
    gene_namesRe: selected genes in the genomaps       
    """

    nump=rowNum*colNum
    if nump<data.shape[1]:
        data,index=select_n_features(data,nump)
        gene_namesRe=gene_names[index]
      #  X_train=X_train[top_n_features]
    #ldaOutzc=scipy.stats.zscore(data, axis=0, ddof=1)    
    genoMaps,T=construct_genomap_returnT(data,rowNum,colNum,epsilon=0.0,num_iter=200)
    return genoMaps,gene_namesRe,T


# -*- coding: utf-8 -*-
"""
Created on Sun May 15 09:03:29 2022
Unsupervised genoNet
@author: anonymous
"""

import os
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
import torch.nn.functional as F

torch.manual_seed(0)
torch.use_deterministic_algorithms(True)
def findConv2dOutShape(hin,win,conv,pool=2):
    # get conv arguments
    kernel_size=conv.kernel_size
    stride=conv.stride
    padding=conv.padding
    dilation=conv.dilation

    hout=np.floor((hin+2*padding[0]-dilation[0]*(kernel_size[0]-1)-1)/stride[0]+1)
    wout=np.floor((win+2*padding[1]-dilation[1]*(kernel_size[1]-1)-1)/stride[1]+1)

    if pool:
        hout/=pool
        wout/=pool
    return int(hout),int(wout)


class genoNet(nn.Module):
    # Define the convolutional neural network architecture
    def __init__(self, input_shape, class_num):
        super(genoNet, self).__init__()        
        
        input_dim=input_shape[2]
        Cin, Hin, Win = 1, input_dim, input_dim
        init_f = 8
        num_fc1 = 512        
        self.conv1 = nn.Conv2d(Cin, init_f, kernel_size=3, padding=1)
        h,w=findConv2dOutShape(Hin,Win,self.conv1,pool=0)
        self.num_flatten=h*w*init_f
        self.fc1 = nn.Linear(self.num_flatten, num_fc1)
        self.fc2 = nn.Linear(num_fc1, class_num)
        self.dropout = nn.Dropout(0.25)   
        self.softmax = nn.Softmax(dim=-1)
        
 
    def forward(self, x):
        x = F.relu(self.conv1(x)); 
        x = x.contiguous().view(-1, self.num_flatten)
        x = F.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.fc2(x)
        
        return F.log_softmax(x, dim=1)
    
    def forwardX(self, x):
        x = F.relu(self.conv1(x)); 
        x = x.contiguous().view(-1, self.num_flatten)
        x = F.relu(self.fc1(x))

        return x
    
def _get_device():
    # help function to detect whether the computer has a GPU
    if torch.cuda.is_available():
        device = 'cuda:0'
    else:
        device = 'cpu'
    return device

def fit(model, dataloader, epoch, criterion, optimizer, device, trainset):
    # fit one epoch
    #
    # model: the genoNet model
    # epoch: the current epoch number
    # criterion: the loss function (by default MSE)
    # optimizer: the optimizer (by default Adam)
    # device: the device that the model is running on
    # trainset: the training dataset
    
    model.train()
    running_loss = 0.0
    counter = 0
    
    for i, data in tqdm(enumerate(dataloader),
                        total=int(len(trainset)/dataloader.batch_size), disable=True):
        counter += 1
        img = data[0]
        img = img.to(device).float()
        
        label = data[1]
        label = label.to(device).long()
        optimizer.zero_grad()
        outputs = model(img)

        loss = criterion(outputs, label)
        
        loss.backward()
        optimizer.step()
        running_loss += loss.item()
    
    epoch_loss = running_loss / counter
    return epoch_loss

class geneDataset(Dataset):
    # Dataset for gene expression data
    # 
    # Attribute:
    #  data: rows denote cells and columns denote the genes. 
    def __init__(self,data,label):
        self.data = data
        self.label = label
        
    def __getitem__(self, index):
        return self.data[index,:,:,:], self.label[index]
    
    def __len__(self):
        return len(self.data)

def predict(model, dataloader, device):
    # predict the transformed data to calculate ideal distribution with trained model
    prediction_list = []
    model.eval()
    running_loss = 0.0
    counter = 0
    with torch.no_grad():
        for i, data in enumerate(dataloader):
            img = data[0]
            img = img.to(device).float()

            pred = model(img)
            prediction_list.append(pred.cpu().numpy())
    return np.concatenate(prediction_list)
    
def traingenoNet(data, labels, maxEPOCH, batchSize=16, verbose=False):
    # train the Sparse Autoencoder model
    #
    # data: training data, rows denote cells and columns denote the genes.
    # reduced_Dim: neurons in the hidden layer
    # maxEPOCH: max training EPOCH number
    # batchSize: batchSize for training
    # l2: the weight for l2 regularizer of all parameters
    # sparse_rho: the desired value for KL Divergence
    # sparse: the weight for sparse regularizer (KL Divergence loss)
    device = _get_device()
    
    trainset = geneDataset(data, labels)
    trainloader = DataLoader(trainset,
                             batch_size=batchSize,
                             shuffle=True)
    
    input_dim = data.shape[1:] # 33 x 33 x 1
    class_num, class_weight = np.unique(labels, return_counts=True)
    class_num = len(class_num)
    class_weight = torch.Tensor(np.sum(class_weight) / class_weight).to(device)
    model = genoNet(input_dim, class_num).to(device)    
    criterion = nn.CrossEntropyLoss(weight=class_weight)
    optimizer = optim.AdamW(model.parameters(), lr=1e-3, weight_decay=1e-5)
    train_loss = []
    for epoch in range(maxEPOCH):
        train_epoch_loss = fit(model, trainloader, epoch,
                               criterion, optimizer,device, trainset)
        if verbose:
            if epoch % 10 == 0:
                print(str(epoch) + ': ' + str(train_epoch_loss))
        train_loss.append(train_epoch_loss)
    
    
    return model

def load_genoNet(input_shape, class_num, path):
    device = torch.device('cpu')
    model = genoNet(input_shape, class_num)
    model.load_state_dict(torch.load(path, map_location=device))
    model.eval()
    return model


"""
Tensorflow implementation for ConvIDEC (also known as DCEC) and ConvIDEC-DA algorithms:
    - Xifeng Guo, Xinwang Liu, En Zhu, and Jianping Yin.Deep Clustering with Convolutional Autoencoders. ICONIP 2017.
    - Xifeng Guo, En Zhu, Xinwang Liu, and Jianping Yin. Deep Embedded Clustering with Data Augmentation. ACML 2018.

Author:
    Xifeng Guo. 2018.6.30
"""

from tensorflow.keras.models import Model
from .ConvDEC import ConvDEC


class ConvIDEC(ConvDEC):
    def __init__(self,
                 input_shape,
                 filters=[32, 64, 128, 10],
                 n_clusters=10):

        super(ConvIDEC, self).__init__(input_shape, filters, n_clusters)
        self.model = Model(inputs=self.autoencoder.input,
                           outputs=[self.model.output, self.autoencoder.output])

    def predict(self, x):  # predict cluster labels using the output of clustering layer
        q = self.model.predict(x, verbose=0)[0]
        
        #q = self.model.predict(x, verbose=0)
        return q

    def compile(self, optimizer='sgd', loss=['kld', 'mse'], loss_weights=[0.1, 1.0]):
        self.model.compile(optimizer=optimizer, loss=loss, loss_weights=loss_weights)

    def train_on_batch(self, x, y, sample_weight=None):
        return self.model.train_on_batch(x, [y, x], sample_weight)[0]
